@@ -7,15 +7,13 @@
     [PlatformWindow] trait used by the generated code and the run-time to change
     aspects of windows on the screen.
 */
-use corelib::component::ComponentRc;
 use corelib::items::PointerEventButton;
-use corelib::layout::Orientation;
 use i_slint_core as corelib;
 
 use corelib::graphics::Point;
 use corelib::input::{KeyEvent, KeyEventType, KeyboardModifiers, MouseEvent};
-use corelib::SharedString;
 use corelib::{window::*, Color};
+use corelib::{Coord, SharedString};
 use std::cell::{Cell, RefCell};
 use std::rc::{Rc, Weak};
 use winit::event::WindowEvent;
@@ -51,10 +49,10 @@ pub trait WinitWindow: PlatformWindow {
             }
 
             if (constraints_horizontal, constraints_vertical) != self.constraints() {
-                let min_width = constraints_horizontal.min.min(constraints_horizontal.max);
-                let min_height = constraints_vertical.min.min(constraints_vertical.max);
-                let max_width = constraints_horizontal.max.max(constraints_horizontal.min);
-                let max_height = constraints_vertical.max.max(constraints_vertical.min);
+                let min_width = constraints_horizontal.min.min(constraints_horizontal.max) as f32;
+                let min_height = constraints_vertical.min.min(constraints_vertical.max) as f32;
+                let max_width = constraints_horizontal.max.max(constraints_horizontal.min) as f32;
+                let max_height = constraints_vertical.max.max(constraints_vertical.min) as f32;
 
                 let sf = self.runtime_window().scale_factor();
 
@@ -63,32 +61,31 @@ pub trait WinitWindow: PlatformWindow {
                 } else {
                     None
                 });
-                winit_window.set_max_inner_size(if max_width < f32::MAX || max_height < f32::MAX {
-                    Some(winit::dpi::PhysicalSize::new(
-                        (max_width * sf).min(65535.),
-                        (max_height * sf).min(65535.),
-                    ))
-                } else {
-                    None
-                });
+                winit_window.set_max_inner_size(
+                    if max_width < i32::MAX as f32 || max_height < i32::MAX as f32 {
+                        Some(winit::dpi::PhysicalSize::new(
+                            (max_width * sf).min(65535.),
+                            (max_height * sf).min(65535.),
+                        ))
+                    } else {
+                        None
+                    },
+                );
+                winit_window.set_resizable(true);
                 self.set_constraints((constraints_horizontal, constraints_vertical));
+                winit_window.set_resizable(min_width < max_width || min_height < max_height);
 
                 #[cfg(target_arch = "wasm32")]
                 {
                     // set_max_inner_size / set_min_inner_size don't work on wasm, so apply the size manually
-                    let existing_size = winit_window.inner_size();
-                    if !(min_width..=max_width).contains(&(existing_size.width as f32))
-                        || !(min_height..=max_height).contains(&(existing_size.height as f32))
+                    let existing_size: winit::dpi::LogicalSize<f32> =
+                        winit_window.inner_size().to_logical(sf as f64);
+                    if !(min_width..=max_width).contains(&(existing_size.width))
+                        || !(min_height..=max_height).contains(&(existing_size.height))
                     {
                         let new_size = winit::dpi::LogicalSize::new(
-                            existing_size
-                                .width
-                                .min(max_width.ceil() as u32)
-                                .max(min_width.ceil() as u32),
-                            existing_size
-                                .height
-                                .min(max_height.ceil() as u32)
-                                .max(min_height.ceil() as u32),
+                            existing_size.width.min(max_width).max(min_width),
+                            existing_size.height.min(max_height).max(min_height),
                         );
                         winit_window.set_inner_size(new_size);
                     }
@@ -105,13 +102,14 @@ pub trait WinitWindow: PlatformWindow {
         let title = window_item.title();
         let no_frame = window_item.no_frame();
         let icon = window_item.icon();
-        let width = window_item.width();
-        let height = window_item.height();
+        let mut width = window_item.width() as f32;
+        let mut height = window_item.height() as f32;
 
         self.set_background_color(background);
         self.set_icon(icon);
 
-        let mut size: winit::dpi::LogicalSize<f64> = Default::default();
+        let mut must_resize = false;
+
         self.with_window_handle(&mut |winit_window| {
             winit_window.set_title(&title);
             if no_frame && winit_window.fullscreen().is_none() {
@@ -119,55 +117,40 @@ pub trait WinitWindow: PlatformWindow {
             } else {
                 winit_window.set_decorations(true);
             }
-            size =
-                winit_window.inner_size().to_logical(self.runtime_window().scale_factor() as f64);
-        });
 
-        let mut must_resize = false;
-        let mut resizable = None;
-        let mut w = width;
-        let mut h = height;
-        if (size.width as f32 - w).abs() < 1. || (size.height as f32 - h).abs() < 1. {
-            return;
-        }
-        if w <= 0. || h <= 0. {
-            if let Some(component_rc) = self.runtime_window().try_component() {
-                let component = ComponentRc::borrow_pin(&component_rc);
-                let hor_info = component.as_ref().layout_info(Orientation::Horizontal);
-                if w <= 0. {
-                    w = hor_info.preferred_bounded();
-                    must_resize = true;
-                }
-                let ver_info = component.as_ref().layout_info(Orientation::Vertical);
-                if h <= 0. {
-                    h = ver_info.preferred_bounded();
-                    must_resize = true;
-                }
-                resizable = Some(hor_info.min != hor_info.max && ver_info.min != ver_info.max);
+            let existing_size =
+                winit_window.inner_size().to_logical(self.runtime_window().scale_factor() as f64);
+
+            if width <= 0. {
+                width = existing_size.width;
+                must_resize = true;
             }
-        };
-        if w > 0. {
-            size.width = w as _;
-        }
-        if h > 0. {
-            size.height = h as _;
-        }
-        self.with_window_handle(&mut |winit_window| {
-            // If we're in fullscreen state, don't try to resize the window but maintain the surface
-            // size we've been assigned to from the windowing system. Weston/Wayland don't like it
-            // when we create a surface that's bigger than the screen due to constraints (#532).
-            if winit_window.fullscreen().is_none() {
-                winit_window.set_inner_size(size);
-                if let Some(resizable) = resizable {
-                    winit_window.set_resizable(resizable)
+            if height <= 0. {
+                height = existing_size.height;
+                must_resize = true;
+            }
+
+            if (existing_size.width as f32 - width).abs() > 1.
+                || (existing_size.height as f32 - height).abs() > 1.
+            {
+                // If we're in fullscreen state, don't try to resize the window but maintain the surface
+                // size we've been assigned to from the windowing system. Weston/Wayland don't like it
+                // when we create a surface that's bigger than the screen due to constraints (#532).
+                if winit_window.fullscreen().is_none() {
+                    winit_window.set_inner_size(winit::dpi::LogicalSize::new(width, height));
                 }
             }
         });
 
         if must_resize {
-            self.runtime_window().set_window_item_geometry(size.width as _, size.height as _)
+            self.runtime_window().set_window_item_geometry(width as _, height as _)
         }
     }
+
+    /// Return true if the proxy element used for input method has the focus
+    fn input_method_focused(&self) -> bool {
+        false
+    }
 }
 
 struct NotRunningEventLoop {
@@ -300,8 +283,14 @@ pub enum CustomEvent {
     /// web for example to request an animation frame.
     #[cfg(target_arch = "wasm32")]
     RedrawAllWindows,
+    /// On wasm request_redraw doesn't wake the event loop, so we need to manually send an event
+    /// so that the event loop can run
+    #[cfg(target_arch = "wasm32")]
+    WakeEventLoopWorkaround,
     UpdateWindowProperties(winit::window::WindowId),
     UserEvent(Box<dyn FnOnce() + Send>),
+    /// Called from `GLWindow::hide` so that we can check if we should quit the event loop
+    WindowHidden,
     Exit,
 }
 
@@ -310,13 +299,17 @@ impl std::fmt::Debug for CustomEvent {
         match self {
             #[cfg(target_arch = "wasm32")]
             Self::RedrawAllWindows => write!(f, "RedrawAllWindows"),
+            #[cfg(target_arch = "wasm32")]
+            Self::WakeEventLoopWorkaround => write!(f, "WakeEventLoopWorkaround"),
             Self::UpdateWindowProperties(e) => write!(f, "UpdateWindowProperties({:?})", e),
             Self::UserEvent(_) => write!(f, "UserEvent"),
+            Self::WindowHidden => write!(f, "WindowHidden"),
             Self::Exit => write!(f, "Exit"),
         }
     }
 }
 
+#[cfg(target_arch = "wasm32")]
 fn redraw_all_windows() {
     let all_windows_weak =
         ALL_WINDOWS.with(|windows| windows.borrow().values().cloned().collect::<Vec<_>>());
@@ -346,8 +339,6 @@ mod key_codes {
 fn process_window_event(
     window: Rc<dyn WinitWindow>,
     event: WindowEvent,
-    quit_behavior: i_slint_core::backend::EventLoopQuitBehavior,
-    control_flow: &mut winit::event_loop::ControlFlow,
     cursor_pos: &mut Point,
     pressed: &mut bool,
 ) {
@@ -375,19 +366,11 @@ fn process_window_event(
             runtime_window.set_window_item_geometry(size.width, size.height);
         }
         WindowEvent::CloseRequested => {
-            window.hide();
-            match quit_behavior {
-                corelib::backend::EventLoopQuitBehavior::QuitOnLastWindowClosed => {
-                    let window_count = ALL_WINDOWS.with(|windows| windows.borrow().len());
-                    if window_count == 0 {
-                        *control_flow = winit::event_loop::ControlFlow::Exit;
-                    }
-                }
-                corelib::backend::EventLoopQuitBehavior::QuitOnlyExplicitly => {}
+            if runtime_window.request_close() {
+                window.hide();
             }
         }
         WindowEvent::ReceivedCharacter(ch) => {
-            corelib::animations::update_animations();
             // On Windows, X11 and Wayland sequences like Ctrl+C will send a ReceivedCharacter after the pressed keyboard input event,
             // with a control character. We choose not to forward those but try to use the current key code instead.
             //
@@ -417,25 +400,39 @@ fn process_window_event(
             runtime_window.process_key_input(&event);
         }
         WindowEvent::Focused(have_focus) => {
+            let have_focus = have_focus || window.input_method_focused();
             // We don't render popups as separate windows yet, so treat
             // focus to be the same as being active.
-            runtime_window.set_active(have_focus);
-            runtime_window.set_focus(have_focus);
+            if have_focus != runtime_window.active() {
+                runtime_window.set_active(have_focus);
+                runtime_window.set_focus(have_focus);
+            }
         }
         WindowEvent::KeyboardInput { ref input, .. } => {
-            corelib::animations::update_animations();
             window.currently_pressed_key_code().set(match input.state {
                 winit::event::ElementState::Pressed => input.virtual_keycode,
                 _ => None,
             });
             if let Some(text) = input.virtual_keycode.and_then(key_codes::winit_key_to_string) {
+                #[allow(unused_mut)]
+                let mut modifiers = window.current_keyboard_modifiers().get();
+                // On wasm, the WindowEvent::ModifiersChanged event is not received
+                #[cfg(target_arch = "wasm32")]
+                #[allow(deprecated)]
+                {
+                    modifiers.shift |= input.modifiers.shift();
+                    modifiers.control |= input.modifiers.ctrl();
+                    modifiers.meta |= input.modifiers.logo();
+                    modifiers.alt |= input.modifiers.alt();
+                }
+
                 let event = key_event(
                     match input.state {
                         winit::event::ElementState::Pressed => KeyEventType::KeyPressed,
                         winit::event::ElementState::Released => KeyEventType::KeyReleased,
                     },
                     text,
-                    window.current_keyboard_modifiers().get(),
+                    modifiers,
                 );
                 runtime_window.process_key_input(&event);
             };
@@ -452,7 +449,6 @@ fn process_window_event(
             window.current_keyboard_modifiers().set(modifiers);
         }
         WindowEvent::CursorMoved { position, .. } => {
-            corelib::animations::update_animations();
             let position = position.to_logical(runtime_window.scale_factor() as f64);
             *cursor_pos = euclid::point2(position.x, position.y);
             runtime_window.process_mouse_input(MouseEvent::MouseMoved { pos: *cursor_pos });
@@ -460,13 +456,11 @@ fn process_window_event(
         WindowEvent::CursorLeft { .. } => {
             // On the html canvas, we don't get the mouse move or release event when outside the canvas. So we have no choice but canceling the event
             if cfg!(target_arch = "wasm32") || !*pressed {
-                corelib::animations::update_animations();
                 *pressed = false;
                 runtime_window.process_mouse_input(MouseEvent::MouseExit);
             }
         }
         WindowEvent::MouseWheel { delta, .. } => {
-            corelib::animations::update_animations();
             let delta = match delta {
                 winit::event::MouseScrollDelta::LineDelta(lx, ly) => {
                     euclid::point2(lx * 60., ly * 60.)
@@ -475,11 +469,11 @@ fn process_window_event(
                     let d = d.to_logical(runtime_window.scale_factor() as f64);
                     euclid::point2(d.x, d.y)
                 }
-            };
+            }
+            .cast::<Coord>();
             runtime_window.process_mouse_input(MouseEvent::MouseWheel { pos: *cursor_pos, delta });
         }
         WindowEvent::MouseInput { state, button, .. } => {
-            corelib::animations::update_animations();
             let button = match button {
                 winit::event::MouseButton::Left => PointerEventButton::left,
                 winit::event::MouseButton::Right => PointerEventButton::right,
@@ -499,7 +493,6 @@ fn process_window_event(
             runtime_window.process_mouse_input(ev);
         }
         WindowEvent::Touch(touch) => {
-            corelib::animations::update_animations();
             let location = touch.location.to_logical(runtime_window.scale_factor() as f64);
             let pos = euclid::point2(location.x, location.y);
             let ev = match touch.phase {
@@ -564,19 +557,11 @@ pub fn run(quit_behavior: i_slint_core::backend::EventLoopQuitBehavior) {
             match event {
                 winit::event::Event::WindowEvent { event, window_id } => {
                     if let Some(window) = window_by_id(window_id) {
-                        process_window_event(
-                            window,
-                            event,
-                            quit_behavior,
-                            control_flow,
-                            &mut cursor_pos,
-                            &mut pressed,
-                        );
+                        process_window_event(window, event, &mut cursor_pos, &mut pressed);
                     };
                 }
 
                 winit::event::Event::RedrawRequested(id) => {
-                    corelib::animations::update_animations();
                     if let Some(window) = window_by_id(id) {
                         window.draw();
                     }
@@ -587,6 +572,15 @@ pub fn run(quit_behavior: i_slint_core::backend::EventLoopQuitBehavior) {
                         window.runtime_window().update_window_properties();
                     }
                 }
+                winit::event::Event::UserEvent(CustomEvent::WindowHidden) => match quit_behavior {
+                    corelib::backend::EventLoopQuitBehavior::QuitOnLastWindowClosed => {
+                        let window_count = ALL_WINDOWS.with(|windows| windows.borrow().len());
+                        if window_count == 0 {
+                            *control_flow = winit::event_loop::ControlFlow::Exit;
+                        }
+                    }
+                    corelib::backend::EventLoopQuitBehavior::QuitOnlyExplicitly => {}
+                },
 
                 winit::event::Event::UserEvent(CustomEvent::Exit) => {
                     *control_flow = winit::event_loop::ControlFlow::Exit;
@@ -600,6 +594,16 @@ pub fn run(quit_behavior: i_slint_core::backend::EventLoopQuitBehavior) {
                 winit::event::Event::UserEvent(CustomEvent::RedrawAllWindows) => {
                     redraw_all_windows()
                 }
+
+                #[cfg(target_arch = "wasm32")]
+                winit::event::Event::UserEvent(CustomEvent::WakeEventLoopWorkaround) => {
+                    *control_flow = winit::event_loop::ControlFlow::Poll;
+                }
+
+                winit::event::Event::MainEventsCleared => {
+                    corelib::timers::TimerList::maybe_activate_timers();
+                    corelib::animations::update_animations();
+                }
                 _ => (),
             }
 
@@ -608,11 +612,8 @@ pub fn run(quit_behavior: i_slint_core::backend::EventLoopQuitBehavior) {
                     .with(|driver| driver.has_active_animations())
             {
                 *control_flow = ControlFlow::Poll;
-                redraw_all_windows()
             }
 
-            corelib::timers::TimerList::maybe_activate_timers();
-
             if *control_flow == winit::event_loop::ControlFlow::Wait {
                 if let Some(next_timer) = corelib::timers::TimerList::next_timeout() {
                     *control_flow = winit::event_loop::ControlFlow::WaitUntil(next_timer.into());
