@@ -9,23 +9,23 @@ use core::cell::{Cell, RefCell};
 use core::pin::Pin;
 use std::rc::{Rc, Weak};
 
-use super::{ItemGraphicsCache, TextureCache};
+use super::TextureCache;
 use crate::event_loop::WinitWindow;
 use crate::glcontext::OpenGLContext;
+use crate::glrenderer::{CanvasRc, ItemGraphicsCache};
 use const_field_offset::FieldOffsets;
 use corelib::api::{GraphicsAPI, RenderingNotifier, RenderingState, SetRenderingNotifierError};
 use corelib::component::ComponentRc;
-use corelib::graphics::*;
+use corelib::graphics::rendering_metrics_collector::RenderingMetricsCollector;
 use corelib::input::KeyboardModifiers;
 use corelib::items::{ItemRef, MouseCursor};
 use corelib::layout::Orientation;
 use corelib::window::{PlatformWindow, PopupWindow, PopupWindowLocation};
 use corelib::Property;
+use corelib::{graphics::*, Coord};
 use i_slint_core as corelib;
 use winit::dpi::LogicalSize;
 
-use crate::CanvasRc;
-
 pub const PASSWORD_CHARACTER: &str = "●";
 
 /// GraphicsWindow is an implementation of the [PlatformWindow][`crate::eventloop::PlatformWindow`] trait. This is
@@ -40,12 +40,15 @@ pub struct GLWindow {
     // This cache only contains textures. The cache for decoded CPU side images is in crate::IMAGE_CACHE.
     pub(crate) texture_cache: RefCell<TextureCache>,
 
-    fps_counter: Option<Rc<FPSCounter>>,
+    rendering_metrics_collector: Option<Rc<RenderingMetricsCollector>>,
 
     rendering_notifier: RefCell<Option<Box<dyn RenderingNotifier>>>,
 
     #[cfg(target_arch = "wasm32")]
     canvas_id: String,
+
+    #[cfg(target_arch = "wasm32")]
+    virtual_keyboard_helper: RefCell<Option<super::wasm_input_helper::WasmInputHelper>>,
 }
 
 impl GLWindow {
@@ -66,10 +69,12 @@ impl GLWindow {
             currently_pressed_key_code: Default::default(),
             graphics_cache: Default::default(),
             texture_cache: Default::default(),
-            fps_counter: FPSCounter::new(),
+            rendering_metrics_collector: RenderingMetricsCollector::new(window_weak.clone()),
             rendering_notifier: Default::default(),
             #[cfg(target_arch = "wasm32")]
             canvas_id,
+            #[cfg(target_arch = "wasm32")]
+            virtual_keyboard_helper: Default::default(),
         })
     }
 
@@ -188,7 +193,7 @@ impl WinitWindow for GLWindow {
                     0,
                     size.width,
                     size.height,
-                    crate::to_femtovg_color(&window.clear_color),
+                    crate::glrenderer::to_femtovg_color(&window.clear_color),
                 );
                 // For the BeforeRendering rendering notifier callback it's important that this happens *after* clearing
                 // the back buffer, in order to allow the callback to provide its own rendering of the background.
@@ -204,27 +209,19 @@ impl WinitWindow for GLWindow {
                 }
             }
 
-            let mut renderer = crate::GLItemRenderer {
-                canvas: window.canvas.as_ref().unwrap().clone(),
-                layer_images_to_delete_after_flush: Default::default(),
-                graphics_window: self.clone(),
+            let mut renderer = crate::glrenderer::GLItemRenderer::new(
+                window.canvas.as_ref().unwrap().clone(),
+                self.clone(),
                 scale_factor,
-                state: vec![crate::State {
-                    scissor: Rect::new(
-                        Point::default(),
-                        Size::new(size.width as f32, size.height as f32) / scale_factor,
-                    ),
-                    global_alpha: 1.,
-                    layer: None,
-                }],
-            };
+                size,
+            );
 
             for (component, origin) in components {
                 corelib::item_rendering::render_component_items(component, &mut renderer, *origin);
             }
 
-            if let Some(fps_counter) = &self.fps_counter {
-                fps_counter.measure_frame_rendered(&mut renderer);
+            if let Some(collector) = &self.rendering_metrics_collector {
+                collector.measure_frame_rendered(&mut renderer);
             }
 
             renderer.canvas.borrow_mut().flush();
@@ -280,6 +277,16 @@ impl WinitWindow for GLWindow {
             }
         };
     }
+
+    #[cfg(target_arch = "wasm32")]
+    fn input_method_focused(&self) -> bool {
+        match self.virtual_keyboard_helper.try_borrow() {
+            Ok(vkh) => vkh.as_ref().map_or(false, |h| h.has_focus()),
+            // the only location in which the virtual_keyboard_helper is mutably borrowed is from
+            // show_virtual_keyboard, which means we have the focus
+            Err(_) => true,
+        }
+    }
 }
 
 impl PlatformWindow for GLWindow {
@@ -395,7 +402,7 @@ impl PlatformWindow for GLWindow {
             (
                 window_item.title().to_string(),
                 window_item.no_frame(),
-                window_item.height() == 0. && window_item.width() == 0.,
+                window_item.height() <= 0 as _ && window_item.width() <= 0 as _,
             )
         } else {
             ("Slint Window".to_string(), false, true)
@@ -405,10 +412,17 @@ impl PlatformWindow for GLWindow {
             .with_title(window_title)
             .with_resizable(is_resizable);
 
-        let scale_factor_override = std::env::var("SLINT_SCALE_FACTOR")
-            .ok()
-            .and_then(|x| x.parse::<f64>().ok())
-            .filter(|f| *f > 0.);
+        let scale_factor_override = runtime_window.scale_factor();
+        // If the scale factor was already set programmatically, use that
+        // else, use the SLINT_SCALE_FACTOR if set, otherwise use the one from winit
+        let scale_factor_override = if scale_factor_override > 1. {
+            Some(scale_factor_override as f64)
+        } else {
+            std::env::var("SLINT_SCALE_FACTOR")
+                .ok()
+                .and_then(|x| x.parse::<f64>().ok())
+                .filter(|f| *f > 0.)
+        };
 
         let window_builder = if std::env::var("SLINT_FULLSCREEN").is_ok() {
             window_builder.with_fullscreen(Some(winit::window::Fullscreen::Borderless(None)))
@@ -420,7 +434,7 @@ impl PlatformWindow for GLWindow {
                 layout_info_v.preferred_bounded(),
             );
 
-            if s.width > 0. && s.height > 0. {
+            if s.width > 0 as Coord && s.height > 0 as Coord {
                 // Make sure that the window's inner size is in sync with the root window item's
                 // width/height.
                 runtime_window.set_window_item_geometry(s.width, s.height);
@@ -463,7 +477,7 @@ impl PlatformWindow for GLWindow {
         );
         let id = platform_window.id();
 
-        if let Some(fps_counter) = &self.fps_counter {
+        if let Some(collector) = &self.rendering_metrics_collector {
             cfg_if::cfg_if! {
                 if #[cfg(target_arch = "wasm32")] {
                     let winsys = "HTML Canvas";
@@ -495,7 +509,7 @@ impl PlatformWindow for GLWindow {
                 }
             }
 
-            fps_counter.start(&format!("GL backend (windowing system: {})", winsys));
+            collector.start(&format!("GL backend (windowing system: {})", winsys));
         }
 
         drop(platform_window);
@@ -519,6 +533,10 @@ impl PlatformWindow for GLWindow {
         if let Some(existing_blinker) = self.cursor_blinker.borrow().upgrade() {
             existing_blinker.stop();
         }*/
+        crate::event_loop::with_window_target(|event_loop| {
+            event_loop.event_loop_proxy().send_event(crate::event_loop::CustomEvent::WindowHidden)
+        })
+        .unwrap();
     }
 
     fn set_mouse_cursor(&self, cursor: MouseCursor) {
@@ -533,6 +551,7 @@ impl PlatformWindow for GLWindow {
             MouseCursor::text => winit::window::CursorIcon::Text,
             MouseCursor::alias => winit::window::CursorIcon::Alias,
             MouseCursor::copy => winit::window::CursorIcon::Copy,
+            MouseCursor::r#move => winit::window::CursorIcon::Move,
             MouseCursor::no_drop => winit::window::CursorIcon::NoDrop,
             MouseCursor::not_allowed => winit::window::CursorIcon::NotAllowed,
             MouseCursor::grab => winit::window::CursorIcon::Grab,
@@ -562,7 +581,7 @@ impl PlatformWindow for GLWindow {
         &self,
         font_request: corelib::graphics::FontRequest,
         text: &str,
-        max_width: Option<f32>,
+        max_width: Option<Coord>,
     ) -> Size {
         let font_request = font_request.merge(&self.default_font_properties());
 
@@ -587,7 +606,7 @@ impl PlatformWindow for GLWindow {
 
         let width = text_input.width() * scale_factor;
         let height = text_input.height() * scale_factor;
-        if width <= 0. || height <= 0. {
+        if width <= 0. || height <= 0. || pos.y < 0. {
             return 0;
         }
 
@@ -645,20 +664,26 @@ impl PlatformWindow for GLWindow {
         }
     }
 
-    fn text_input_position_for_byte_offset(
+    fn text_input_cursor_rect_for_byte_offset(
         &self,
         text_input: Pin<&corelib::items::TextInput>,
         byte_offset: usize,
-    ) -> Point {
+    ) -> Rect {
         let scale_factor = self.self_weak.upgrade().unwrap().scale_factor();
         let text = text_input.text();
 
+        let font_size = text_input
+            .unresolved_font_request()
+            .merge(&self.default_font_properties())
+            .pixel_size
+            .unwrap_or(super::fonts::DEFAULT_FONT_SIZE);
+
         let mut result = Point::default();
 
         let width = text_input.width() * scale_factor;
         let height = text_input.height() * scale_factor;
         if width <= 0. || height <= 0. {
-            return result;
+            return Rect::new(result, Size::new(1.0, font_size));
         }
 
         let font = crate::fonts::FONT_CACHE.with(|cache| {
@@ -683,7 +708,7 @@ impl PlatformWindow for GLWindow {
                 if (start..=(start + line_text.len())).contains(&byte_offset) {
                     for glyph in &metrics.glyphs {
                         if glyph.byte_index == (byte_offset - start) {
-                            result = line_pos + euclid::vec2(glyph.x, glyph.y);
+                            result = line_pos + euclid::vec2(glyph.x, 0.0);
                             return;
                         }
                     }
@@ -694,7 +719,25 @@ impl PlatformWindow for GLWindow {
             },
         );
 
-        result / scale_factor
+        Rect::new(result / scale_factor, Size::new(1.0, font_size))
+    }
+
+    #[cfg(target_arch = "wasm32")]
+    fn show_virtual_keyboard(&self, _it: corelib::items::InputType) {
+        let mut vkh = self.virtual_keyboard_helper.borrow_mut();
+        let h = vkh.get_or_insert_with(|| {
+            let canvas =
+                self.borrow_mapped_window().unwrap().opengl_context.html_canvas_element().clone();
+            super::wasm_input_helper::WasmInputHelper::new(self.self_weak.clone(), canvas)
+        });
+        h.show();
+    }
+
+    #[cfg(target_arch = "wasm32")]
+    fn hide_virtual_keyboard(&self) {
+        if let Some(h) = &*self.virtual_keyboard_helper.borrow() {
+            h.hide()
+        }
     }
 
     fn as_any(&self) -> &dyn std::any::Any {
