@@ -10,7 +10,7 @@ use i_slint_core::graphics::{IntSize, SharedImageBuffer};
 use i_slint_core::Property;
 use i_slint_core::{items::ImageRendering, slice::Slice, ImageInner, SharedString};
 
-use super::{CanvasRc, GLItemRenderer};
+use super::glrenderer::{CanvasRc, GLItemRenderer};
 
 struct Texture {
     id: femtovg::ImageId,
@@ -134,7 +134,7 @@ impl std::fmt::Debug for ImageData {
 }
 
 #[derive(Debug)]
-pub(crate) struct CachedImage(RefCell<ImageData>);
+pub struct CachedImage(RefCell<ImageData>);
 
 impl CachedImage {
     fn new_on_cpu(decoded_image: image::DynamicImage, premultiplied_alpha: bool) -> Self {
@@ -424,12 +424,24 @@ impl CachedImage {
     }
 
     pub(crate) fn as_paint(&self) -> femtovg::Paint {
+        self.as_paint_with_alpha(1.0)
+    }
+
+    pub(crate) fn as_paint_with_alpha(&self, alpha_tint: f32) -> femtovg::Paint {
         match &*self.0.borrow() {
             ImageData::Texture(tex) => {
                 let size = tex
                     .size()
                     .expect("internal error: CachedImage::as_paint() called on zero-sized texture");
-                femtovg::Paint::image(tex.id, 0., 0., size.width as f32, size.height as f32, 0., 1.)
+                femtovg::Paint::image(
+                    tex.id,
+                    0.,
+                    0.,
+                    size.width as f32,
+                    size.height as f32,
+                    0.,
+                    alpha_tint,
+                )
             }
             _ => panic!("internal error: CachedImage::as_paint() called on non-texture image"),
         }
